/*
 * Copyright (C) 2013 Giorgio Vazzana
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/* pf-upnp.c: simple cli program to list/query/add/remove port forwarding entries using UPnP
 * depends:   libgupnp-1.0-dev
 * compile:   gcc -Wall -O2 $(pkg-config --cflags gupnp-1.0) -o pf-upnp pf-upnp.c $(pkg-config --libs gupnp-1.0 gssdp-1.0 libxml-2.0 libsoup-2.4)
 * examples:
 *  1) get status info and external IP address:
 *     ./pf-upnp
 *  2) list port mappings:
 *     ./pf-upnp -l
 *  3) query mapping (udp port 8110):
 *     ./pf-upnp -q -u -p 8110
 *  4) add port mapping (open udp port 8110):
 *     ./pf-upnp -a -u -p 8110 -c 192.168.1.7 -D parole
 *  5) delete mapping (close udp port 8110):
 *     ./pf-upnp -d -u -p 8110
 */

#include <libgupnp/gupnp-control-point.h>

#define VERSION "0.0.1"
#define DEBUG 0

struct status_info_st {
	gchar *connection_status;
	gchar *last_connection_error;
	guint  uptime;
};

struct port_mapping_st {
	gchar    *remote_host;
	guint     external_port;
	gchar    *protocol;
	guint     internal_port;
	gchar    *internal_client;
	gboolean  enabled;
	gchar    *description;
	guint     lease_duration;
};

struct operations_st {
	gboolean                list_mappings;
	gboolean                query_mapping;
	gboolean                add_mapping;
	gboolean                del_mapping;
	struct port_mapping_st *pm;
	gint                    exit_code;
};

static GMainLoop *main_loop;
static gboolean   opt_list_mappings   = FALSE;
static gboolean   opt_query_mapping   = FALSE;
static gboolean   opt_add_mapping     = FALSE;
static gboolean   opt_del_mapping     = FALSE;
static guint      opt_port            = 0;
static gboolean   opt_udp             = FALSE;
static gchar     *opt_internal_client = NULL;
static gchar     *opt_description     = NULL;
static guint      opt_timeout         = 5;

static GOptionEntry entries[] =
{
	{ "list", 'l', 0, G_OPTION_ARG_NONE, &opt_list_mappings, "List port mappings", NULL },
	{ "query", 'q', 0, G_OPTION_ARG_NONE, &opt_query_mapping, "Query mapping", NULL },
	{ "add", 'a', 0, G_OPTION_ARG_NONE, &opt_add_mapping, "Add mapping", NULL },
	{ "del", 'd', 0, G_OPTION_ARG_NONE, &opt_del_mapping, "Delete mapping", NULL },
	{ "port", 'p', 0, G_OPTION_ARG_INT, &opt_port, "Port number", "N" },
	{ "udp", 'u', 0, G_OPTION_ARG_NONE, &opt_udp, "udp port (tcp otherwise)", NULL },
	{ "client", 'c', 0, G_OPTION_ARG_STRING, &opt_internal_client, "Internal client", "IP" },
	{ "desc", 'D', 0, G_OPTION_ARG_STRING, &opt_description, "Description", "STRING" },
	{ "timeout", 't', 0, G_OPTION_ARG_INT, &opt_timeout, "Timeout (default 5s)", "N" },
	{ NULL }
};

static gboolean timeout_cb(gpointer data)
{
	struct operations_st *op = data;

	op->exit_code = 2;
	g_printerr("Error: Timeout\n");
	g_main_loop_quit(main_loop);

	return FALSE;
}

static void service_proxy_available_cb(GUPnPControlPoint *cp, GUPnPServiceProxy *proxy, gpointer data)
{
	struct operations_st   *op = data;
	GError                 *error;
	gchar                  *ip;
	struct status_info_st  *si;
	struct port_mapping_st *pm;

	/* get connections status */
	error = NULL;
	si = g_malloc(sizeof(*si));
	gupnp_service_proxy_send_action(proxy,
		/* Action name and error location */
		"GetStatusInfo", &error,
		/* IN args */
		NULL,
		/* OUT args */
		"NewConnectionStatus",
		G_TYPE_STRING, &si->connection_status,
		"NewLastConnectionError",
		G_TYPE_STRING, &si->last_connection_error,
		"NewUptime",
		G_TYPE_UINT, &si->uptime,
		NULL);

	if (error == NULL) {
		g_print("Connection Status:   %s\n", si->connection_status);
		g_print("Uptime:              %uh%02um%02us\n", si->uptime/3600, (si->uptime/60)%60, si->uptime%60);
		g_free(si->connection_status);
		g_free(si->last_connection_error);
		g_free(si);
	} else {
		g_printerr("Error: GetStatusInfo: %s (%u)\n", error->message, error->code);
		g_error_free(error);
		goto fail;
	}

	/* get external ip address */
	error = NULL;
	ip    = NULL;
	gupnp_service_proxy_send_action(proxy,
		/* Action name and error location */
		"GetExternalIPAddress", &error,
		/* IN args */
		NULL,
		/* OUT args */
		"NewExternalIPAddress",
		G_TYPE_STRING, &ip,
		NULL);

	if (error == NULL) {
		g_print("External IP address: %s\n", ip);
		g_free(ip);
	} else {
		g_printerr("Error: GetExternalIPAddress: %s (%u)\n", error->message, error->code);
		g_error_free(error);
		goto fail;
	}

	/* query mapping */
	if (op->query_mapping == TRUE) {
		struct port_mapping_st *pm = op->pm;

		g_print("Querying port mapping...\n");
		error = NULL;
		gupnp_service_proxy_send_action(proxy,
			/* Action name and error location */
			"GetSpecificPortMappingEntry", &error,
			/* IN args */
			"NewRemoteHost",
			G_TYPE_STRING, pm->remote_host,
			"NewExternalPort",
			G_TYPE_UINT, pm->external_port,
			"NewProtocol",
			G_TYPE_STRING, pm->protocol,
			NULL,
			/* OUT args */
			"NewInternalPort",
			G_TYPE_UINT, &pm->internal_port,
			"NewInternalClient",
			G_TYPE_STRING, &pm->internal_client,
			"NewEnabled",
			G_TYPE_BOOLEAN, &pm->enabled,
			"NewPortMappingDescription",
			G_TYPE_STRING, &pm->description,
			"NewLeaseDuration",
			G_TYPE_UINT, &pm->lease_duration,
			NULL);

		if (error == NULL) {
			g_print("  Query result:\n");
			g_print("    Remote host:     %s\n", pm->remote_host);
			g_print("    External port:   %u\n", pm->external_port);
			g_print("    Protocol:        %s\n", pm->protocol);
			g_print("    Internal port:   %u\n", pm->internal_port);
			g_print("    Internal client: %s\n", pm->internal_client);
			g_print("    Enabled:         %u\n", pm->enabled);
			g_print("    Description:     %s\n", pm->description);
			g_print("    Lease duration:  %u\n", pm->lease_duration);
		} else {
			g_printerr("Error: GetSpecificPortMappingEntry: %s (%u)\n", error->message, error->code);
			g_error_free(error);
			goto fail;
		}
	}

	/* add mapping */
	if (op->add_mapping == TRUE) {
		struct port_mapping_st *pm = op->pm;

		g_print("Adding port mapping...\n");
		error = NULL;
		gupnp_service_proxy_send_action(proxy,
			/* Action name and error location */
			"AddPortMapping", &error,
			/* IN args */
			"NewRemoteHost",
			G_TYPE_STRING, pm->remote_host,
			"NewExternalPort",
			G_TYPE_UINT, pm->external_port,
			"NewProtocol",
			G_TYPE_STRING, pm->protocol,
			"NewInternalPort",
			G_TYPE_UINT, pm->internal_port,
			"NewInternalClient",
			G_TYPE_STRING, pm->internal_client,
			"NewEnabled",
			G_TYPE_BOOLEAN, pm->enabled,
			"NewPortMappingDescription",
			G_TYPE_STRING, pm->description,
			"NewLeaseDuration",
			G_TYPE_UINT, pm->lease_duration,
			NULL,
			/* OUT args */
			NULL);

		if (error == NULL) {
			g_print("  Entry added:\n");
			g_print("    Remote host:     %s\n", pm->remote_host);
			g_print("    External port:   %u\n", pm->external_port);
			g_print("    Protocol:        %s\n", pm->protocol);
			g_print("    Internal port:   %u\n", pm->internal_port);
			g_print("    Internal client: %s\n", pm->internal_client);
			g_print("    Enabled:         %u\n", pm->enabled);
			g_print("    Description:     %s\n", pm->description);
			g_print("    Lease duration:  %u\n", pm->lease_duration);
		} else {
			g_printerr("Error: AddPortMapping: %s (%u)\n", error->message, error->code);
			g_error_free(error);
			goto fail;
		}
	}

	/* delete mapping */
	if (op->del_mapping == TRUE) {
		struct port_mapping_st *pm = op->pm;

		g_print("Deleting port mapping...\n");
		error = NULL;
		gupnp_service_proxy_send_action(proxy,
			/* Action name and error location */
			"DeletePortMapping", &error,
			/* IN args */
			"NewRemoteHost",
			G_TYPE_STRING, pm->remote_host,
			"NewExternalPort",
			G_TYPE_UINT, pm->external_port,
			"NewProtocol",
			G_TYPE_STRING, pm->protocol,
			NULL,
			/* OUT args */
			NULL);

		if (error == NULL) {
			g_print("  Entry deleted:\n");
			g_print("    Remote host:     %s\n", pm->remote_host);
			g_print("    External port:   %u\n", pm->external_port);
			g_print("    Protocol:        %s\n", pm->protocol);
		} else {
			g_printerr("Error: DeletePortMapping: %s (%u)\n", error->message, error->code);
			g_error_free(error);
			goto fail;
		}
	}

	/* list mappings */
	if (op->list_mappings == TRUE) {
		guint index;

		g_print("Listing port mappings...\n");
		pm = NULL;
		index = 0;
		while (1) {
			error = NULL;
			pm = g_malloc(sizeof(*pm));

			gupnp_service_proxy_send_action(proxy,
				/* Action name and error location */
				"GetGenericPortMappingEntry", &error,
				/* IN args */
				"NewPortMappingIndex",
				G_TYPE_UINT, index,
				NULL,
				/* OUT args */
				"NewRemoteHost",
				G_TYPE_STRING, &pm->remote_host,
				"NewExternalPort",
				G_TYPE_UINT, &pm->external_port,
				"NewProtocol",
				G_TYPE_STRING, &pm->protocol,
				"NewInternalPort",
				G_TYPE_UINT, &pm->internal_port,
				"NewInternalClient",
				G_TYPE_STRING, &pm->internal_client,
				"NewEnabled",
				G_TYPE_BOOLEAN, &pm->enabled,
				"NewPortMappingDescription",
				G_TYPE_STRING, &pm->description,
				"NewLeaseDuration",
				G_TYPE_UINT, &pm->lease_duration,
				NULL);

			if (error == NULL) {
				g_print("  Index: %u\n", index);
				g_print("    Remote host:     %s\n", pm->remote_host);
				g_print("    External port:   %u\n", pm->external_port);
				g_print("    Protocol:        %s\n", pm->protocol);
				g_print("    Internal port:   %u\n", pm->internal_port);
				g_print("    Internal client: %s\n", pm->internal_client);
				g_print("    Enabled:         %u\n", pm->enabled);
				g_print("    Description:     %s\n", pm->description);
				g_print("    Lease duration:  %u\n", pm->lease_duration);
				g_free(pm->remote_host);
				g_free(pm->protocol);
				g_free(pm->internal_client);
				g_free(pm->description);
				g_free(pm);
				pm = NULL;
			} else {
				if (error->code != 713) { // end of ports array
					g_printerr("Error: GetGenericPortMappingEntry: %s (%u)\n", error->message, error->code);
					g_error_free(error);
					g_free(pm);
					goto fail;
				}
				g_error_free(error);
				g_free(pm);
				break;
			}
			index++;
		}
	}

	g_main_loop_quit(main_loop);
	return;

fail:
	op->exit_code = 1;
	g_main_loop_quit(main_loop);
}

static void list_services(GUPnPControlPoint *cp, GUPnPDeviceInfo *info, gpointer data, guint level)
{
	GList *services, *l;
	const char *service_type = NULL;
#if DEBUG
	gchar *tab[4] = {"", "  ", "    ", "      " };
#endif

	/* enum services */
	services = gupnp_device_info_list_services(info);
	l = services;
	while (l) {
		service_type = gupnp_service_info_get_service_type(l->data);
#if DEBUG
		g_print("%sService type:        %s\n", tab[level], service_type);
#endif

		if (g_strcmp0 (service_type, "urn:schemas-upnp-org:service:WANIPConnection:1") == 0 ||
		    g_strcmp0 (service_type, "urn:schemas-upnp-org:service:WANPPPConnection:1") == 0) {
#if 0
			GUPnPControlPoint *cp1;
			cp1 = gupnp_control_point_new(gupnp_control_point_get_context(cp), service_type);
			g_signal_connect(cp1,
			                 "service-proxy-available",
			                  G_CALLBACK(service_proxy_available_cb),
			                  data);
			gssdp_resource_browser_set_active(GSSDP_RESOURCE_BROWSER(cp1), TRUE);
#else
			/* we're casting a GUPnPServiceInfo to a GUPnPServiceProxy, is this possible? */
			service_proxy_available_cb(cp, GUPNP_SERVICE_PROXY(l->data), data);
#endif
		}
		l = l->next;
	}
}

static void device_proxy_available_cb(GUPnPControlPoint *cp, GUPnPDeviceProxy *proxy, gpointer data)
{
	GList *subdev1, *l1, *subdev2, *l2;
	char *friendly_name     = NULL;
#if DEBUG
	const char *device_type = NULL;
#endif

	friendly_name = gupnp_device_info_get_friendly_name(GUPNP_DEVICE_INFO(proxy));
	g_print("Router:              %s\n", friendly_name);
#if DEBUG
	device_type   = gupnp_device_info_get_device_type(GUPNP_DEVICE_INFO(proxy));
	g_print("Device type:         %s\n", device_type);
#endif
	list_services(cp, GUPNP_DEVICE_INFO(proxy), data, 1);
	g_free(friendly_name);

	/* Enum subdevices */
	subdev1 = gupnp_device_info_list_devices(GUPNP_DEVICE_INFO(proxy));
	l1 = subdev1;
	while (l1) {
#if DEBUG
		device_type = gupnp_device_info_get_device_type(l1->data);
		g_print("  Device type:         %s\n", device_type);
#endif
		list_services(cp, l1->data, data, 2);

		subdev2 = gupnp_device_info_list_devices(l1->data);
		l2 = subdev2;
		while (l2) {
#if DEBUG
			device_type = gupnp_device_info_get_device_type(l2->data);
			g_print("    Device type:         %s\n", device_type);
#endif
			list_services(cp, l2->data, data, 3);

			l2 = l2->next;
		}

		l1 = l1->next;
	}
}

int main(int argc, char *argv[])
{
	gint                  i;
	GError               *error;
	GOptionContext       *opt_context;
	GUPnPContext         *upnp_context;
	GUPnPControlPoint    *cp;
	struct operations_st *op;

	/* Parse command line options */
	opt_context = g_option_context_new("- Port Forwarding via UPnP, ver. "VERSION);
	g_option_context_add_main_entries(opt_context, entries, NULL);
	error = NULL;
	if (!g_option_context_parse(opt_context, &argc, &argv, &error)) {
		g_print("option parsing failed: %s\n", error->message);
		exit(1);
	}

	i = 0;
	if (opt_query_mapping == TRUE)
		i++;
	if (opt_add_mapping == TRUE)
		i++;
	if (opt_del_mapping == TRUE)
		i++;
	if (i > 1) {
		g_print("Please specify only one option in {-q, -a,-d}\n");
		exit(1);
	}

	op = g_malloc(sizeof(*op));
	op->list_mappings = opt_list_mappings;
	op->query_mapping = FALSE;
	op->add_mapping   = FALSE;
	op->del_mapping   = FALSE;
	op->pm            = NULL;
	op->exit_code     = 0;
	if (opt_query_mapping == TRUE) {
		if (opt_port) {
			op->query_mapping       = TRUE;
			op->pm                  = g_malloc(sizeof(*op->pm));
			op->pm->remote_host     = g_strdup("");
			op->pm->external_port   = opt_port;
			op->pm->protocol        = (opt_udp == TRUE) ? g_strdup("UDP") : g_strdup("TCP");
		}
	} else if (opt_add_mapping == TRUE) {
		if (opt_port && opt_internal_client && opt_description) {
			op->add_mapping         = TRUE;
			op->pm                  = g_malloc(sizeof(*op->pm));
			op->pm->remote_host     = g_strdup("");
			op->pm->external_port   = opt_port;
			op->pm->protocol        = (opt_udp == TRUE) ? g_strdup("UDP") : g_strdup("TCP");
			op->pm->internal_port   = opt_port;
			op->pm->internal_client = opt_internal_client;
			op->pm->enabled         = TRUE;
			op->pm->description     = opt_description;
			op->pm->lease_duration  = 0;
		}
	} else if (opt_del_mapping == TRUE) {
		if (opt_port) {
			op->del_mapping         = TRUE;
			op->pm                  = g_malloc(sizeof(*op->pm));
			op->pm->remote_host     = g_strdup("");
			op->pm->external_port   = opt_port;
			op->pm->protocol        = (opt_udp == TRUE) ? g_strdup("UDP") : g_strdup("TCP");
		}
	}

	/* Required initialisation */
	g_type_init();

	/* Create a new GUPnP Context. By here we are using the default GLib main
	context, and connecting to the current machine's default IP on an
	automatically generated port. */
	error = NULL;
	upnp_context = gupnp_context_new(NULL, NULL, 0, &error);
	if (error) {
		g_printerr("Error: gupnp_context_new(): %s (%u)\n", error->message, error->code);
		exit(1);
	}

#if 0
	/* Create a Control Point targeting WAN IP Connection services */
	//cp = gupnp_control_point_new(upnp_context, "urn:schemas-upnp-org:service:WANIPConnection:1");
	cp = gupnp_control_point_new(upnp_context, "urn:schemas-upnp-org:service:WANPPPConnection:1");

	/* The service-proxy-available signal is emitted when any services which match
	our target are found, so connect to it */
	g_signal_connect(cp,
	                 "service-proxy-available",
	                 G_CALLBACK(service_proxy_available_cb),
	                 (gpointer)op);
#endif
#if 1
	//cp = gupnp_control_point_new(upnp_context, "upnp:rootdevice");
	cp = gupnp_control_point_new(upnp_context, "urn:schemas-upnp-org:device:InternetGatewayDevice:1");
	g_signal_connect(cp,
	                 "device-proxy-available",
	                 G_CALLBACK(device_proxy_available_cb),
	                 (gpointer)op);
#endif

	/* Tell the Control Point to start searching */
	gssdp_resource_browser_set_active(GSSDP_RESOURCE_BROWSER(cp), TRUE);

	/* add timeout source */
	g_timeout_add_seconds(opt_timeout, timeout_cb, (gpointer)op);

	/* Enter the main loop. This will start the search and result in callbacks to
	service_proxy_available_cb. */
	main_loop = g_main_loop_new(NULL, FALSE);
	g_main_loop_run(main_loop);

	/* Clean up */
	g_main_loop_unref(main_loop);
	g_object_unref(cp);
	g_object_unref(upnp_context);

	return (op->exit_code);
}
